import os
import re
import socket
import time
from base64 import b64encode
from hashlib import sha1
from urlparse import urlparse

from errors import HandshakeError
from python_digest import build_authorization_request


WS_GUID = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
WS_VERSION = '13'
MAX_REDIRECTS = 10
HDR_TIMEOUT = 5
MAX_HDR_LEN = 1024


class Handshake(object):
    def __init__(self, wsock):
        self.wsock = wsock
        self.sock = wsock.sock

    def fail(self, msg):
        self.sock.close()
        raise HandshakeError(msg)

    def receive_request(self):
        raw, headers = self.receive_headers()

        # Request must be HTTP (at least 1.1) GET request, find the location
        # (without trailing slash)
        match = re.search(r'^GET (.*?)/* HTTP/1.1\r\n', raw)

        if match is None:
            self.fail('not a valid HTTP 1.1 GET request')

        location = match.group(1)
        return location, headers

    def receive_response(self):
        raw, headers = self.receive_headers()

        # Response must be HTTP (at least 1.1) with status 101
        match = re.search(r'^HTTP/1\.1 (\d{3})', raw)

        if match is None:
            self.fail('not a valid HTTP 1.1 response')

        status = int(match.group(1))
        return status, headers

    def receive_headers(self):
        # Receive entire HTTP header
        hdr = ''

        sock_timeout = self.sock.gettimeout()

        try:
            force_timeout = sock_timeout is None
            timeout = HDR_TIMEOUT if force_timeout else sock_timeout
            self.sock.settimeout(timeout)

            start_time = time.time()

            while hdr[-4:] != '\r\n\r\n':
                if len(hdr) == MAX_HDR_LEN:
                    raise HandshakeError('request exceeds maximum header '
                                         'length of %d' % MAX_HDR_LEN)

                hdr += self.sock.recv(1)

                time_diff = time.time() - start_time

                if time_diff > timeout:
                    raise socket.timeout

                self.sock.settimeout(timeout - time_diff)
        except socket.timeout:
            self.sock.close()
            raise HandshakeError('timeout while receiving handshake headers')

        self.sock.settimeout(sock_timeout)

        hdr = hdr.decode('utf-8', 'ignore')
        headers = {}

        for key, value in re.findall(r'(.*?): ?(.*?)\r\n', hdr):
            if key in headers:
                headers[key] += ', ' + value
            else:
                headers[key] = value

        return hdr, headers

    def send_headers(self, headers):
        # Send request
        for hdr in list(headers):
            if isinstance(hdr, tuple):
                hdr = '%s: %s' % hdr

            self.sock.sendall(hdr + '\r\n')

        self.sock.sendall('\r\n')

    def perform(self):
        raise NotImplementedError


class ServerHandshake(Handshake):
    """
    Executes a handshake as the server end point of the socket. If the HTTP
    request headers sent by the client are invalid, a HandshakeError is raised.
    """

    def perform(self, ssock):
        # Receive and validate client handshake
        location, headers = self.receive_request()
        self.wsock.location = location
        self.wsock.request_headers = headers

        # Send server handshake in response
        self.send_headers(self.response_headers(ssock))

    def response_headers(self, ssock):
        headers = self.wsock.request_headers

        # Check if headers that MUST be present are actually present
        for name in ('Host', 'Upgrade', 'Connection', 'Sec-WebSocket-Key',
                     'Sec-WebSocket-Version'):
            if name not in headers:
                self.fail('missing "%s" header' % name)

        # Check WebSocket version used by client
        version = headers['Sec-WebSocket-Version']

        if version != WS_VERSION:
            self.fail('WebSocket version %s requested (only %s is supported)'
                      % (version, WS_VERSION))

        # Verify required header keywords
        if 'websocket' not in headers['Upgrade'].lower():
            self.fail('"Upgrade" header must contain "websocket"')

        if 'upgrade' not in headers['Connection'].lower():
            self.fail('"Connection" header must contain "Upgrade"')

        # Origin must be present if browser client, and must match the list of
        # trusted origins
        if 'Origin' not in headers and 'User-Agent' in headers:
            self.fail('browser client must specify "Origin" header')

        origin = headers.get('Origin', 'null')

        if origin == 'null':
            if ssock.trusted_origins:
                self.fail('no "Origin" header specified, assuming untrusted')
        elif ssock.trusted_origins and origin not in ssock.trusted_origins:
            self.fail('untrusted origin "%s"' % origin)

        # Only a supported protocol can be returned
        client_proto = split_stripped(headers['Sec-WebSocket-Protocol']) \
                       if 'Sec-WebSocket-Protocol' in headers else []
        self.wsock.protocol = None

        for p in client_proto:
            if p in ssock.protocols:
                self.wsock.protocol = p
                break

        # Only supported extensions are returned
        if 'Sec-WebSocket-Extensions' in headers:
            self.wsock.extension_instances = []

            for hdr in split_stripped(headers['Sec-WebSocket-Extensions']):
                name, params = parse_param_hdr(hdr)

                for ext in ssock.extensions:
                    if ext.is_supported(name, self.wsock.extension_instances):
                        accept_params = ext.negotiate_safe(name, params)

                        if accept_params is not None:
                            instance = ext.Instance(ext, name, accept_params)
                            self.wsock.extension_instances.append(instance)

        # Check if requested resource location is served by this server
        if ssock.locations:
            if self.wsock.location not in ssock.locations:
                raise HandshakeError('location "%s" is not supported by this '
                                     'server' % self.wsock.location)

        # Encode acceptation key using the WebSocket GUID
        key = headers['Sec-WebSocket-Key'].strip()
        accept = b64encode(sha1(key + WS_GUID).digest())

        # Location scheme differs for SSL-enabled connections
        scheme = 'wss' if self.wsock.secure else 'ws'

        if 'Host' in headers:
            host = headers['Host']
        else:
            host, port = self.sock.getpeername()
            default_port = 443 if self.wsock.secure else 80

            if port != default_port:
                host += ':%d' % port

        location = '%s://%s%s' % (scheme, host, self.wsock.location)

        # Construct HTTP response header
        yield 'HTTP/1.1 101 Switching Protocols'
        yield 'Upgrade', 'websocket'
        yield 'Connection', 'Upgrade'
        yield 'Sec-WebSocket-Origin', origin
        yield 'Sec-WebSocket-Location', location
        yield 'Sec-WebSocket-Accept', accept

        if self.wsock.protocol:
            yield 'Sec-WebSocket-Protocol', self.wsock.protocol

        if self.wsock.extension_instances:
            values = [format_param_hdr(i.name, i.params)
                      for i in self.wsock.extension_instances]
            yield 'Sec-WebSocket-Extensions', ', '.join(values)


class ClientHandshake(Handshake):
    """
    Executes a handshake as the client end point of the socket. May raise a
    HandshakeError if the server response is invalid.
    """

    def __init__(self, wsock):
        Handshake.__init__(self, wsock)
        self.redirects = 0

    def perform(self):
        self.send_headers(self.request_headers())
        self.handle_response(*self.receive_response())

    def handle_response(self, status, headers):
        if status == 101:
            self.handle_handshake(headers)
        elif status == 401:
            self.handle_auth(headers)
        elif status in (301, 302, 303, 307, 308):
            self.handle_redirect(headers)
        else:
            self.fail('invalid HTTP response status %d' % status)

    def handle_handshake(self, headers):
        # Check if headers that MUST be present are actually present
        for name in ('Upgrade', 'Connection', 'Sec-WebSocket-Accept'):
            if name not in headers:
                self.fail('missing "%s" header' % name)

        if 'websocket' not in headers['Upgrade'].lower():
            self.fail('"Upgrade" header must contain "websocket"')

        if 'upgrade' not in headers['Connection'].lower():
            self.fail('"Connection" header must contain "Upgrade"')

        # Verify accept header
        accept = headers['Sec-WebSocket-Accept'].strip()
        required_accept = b64encode(sha1(self.key + WS_GUID).digest())

        if accept != required_accept:
            self.fail('invalid websocket accept header "%s"' % accept)

        # Compare extensions, add hooks only for those returned by server
        if 'Sec-WebSocket-Extensions' in headers:
            # FIXME: there is no distinction between server/client extension
            # instances, while the extension instance may assume it belongs to
            # a server, leading to undefined behavior
            self.wsock.extension_instances = []

            for hdr in split_stripped(headers['Sec-WebSocket-Extensions']):
                name, accept_params = parse_param_hdr(hdr)

                for ext in self.wsock.extensions:
                    if name in ext.names:
                        instance = ext.Instance(ext, name, accept_params)
                        self.wsock.extension_instances.append(instance)
                        break
                else:
                    raise HandshakeError('server handshake contains '
                                         'unsupported extension "%s"' % name)

        # Assert that returned protocol (if any) is supported
        if 'Sec-WebSocket-Protocol' in headers:
            protocol = headers['Sec-WebSocket-Protocol']

            if protocol != 'null' and protocol not in self.wsock.protocols:
                self.fail('unsupported protocol "%s"' % protocol)

            self.wsock.protocol = protocol

    def handle_auth(self, headers):
        # HTTP authentication is required in the request
        hdr = headers['WWW-Authenticate']
        authres = dict(re.findall(r'(\w+)[:=] ?"?(\w+)"?', hdr))
        mode = hdr.lstrip().split(' ', 1)[0]

        if not self.wsock.auth:
            self.fail('missing username and password for HTTP authentication')

        if mode == 'Basic':
            auth_hdr = self.http_auth_basic_headers(**authres)
        elif mode == 'Digest':
            auth_hdr = self.http_auth_digest_headers(**authres)
        else:
            self.fail('unsupported HTTP authentication mode "%s"' % mode)

        # Send new, authenticated handshake
        self.send_headers(list(self.request_headers()) + list(auth_hdr))
        self.handle_response(*self.receive_response())

    def handle_redirect(self, headers):
        self.redirects += 1

        if self.redirects > MAX_REDIRECTS:
            self.fail('reached maximum number of redirects (%d)'
                      % MAX_REDIRECTS)

        # Handle HTTP redirect
        url = urlparse(headers['Location'].strip())

        # Reconnect socket to new host if net location changed
        if not url.port:
            url.port = 443 if self.secure else 80

        addr = (url.netloc, url.port)

        if addr != self.sock.getpeername():
            self.sock.close()
            self.sock.connect(addr)

        # Update websocket object and send new handshake
        self.wsock.location = url.path
        self.perform()

    def request_headers(self):
        if len(self.wsock.location) == 0:
            self.fail('request location is empty')

        # Generate a 16-byte random base64-encoded key for this connection
        self.key = b64encode(os.urandom(16))

        # Send client handshake
        yield 'GET %s HTTP/1.1' % self.wsock.location
        yield 'Host', '%s:%d' % self.sock.getpeername()
        yield 'Upgrade', 'websocket'
        yield 'Connection', 'keep-alive, Upgrade'
        yield 'Sec-WebSocket-Key', self.key
        yield 'Sec-WebSocket-Version', WS_VERSION

        if self.wsock.origin:
            yield 'Origin', self.wsock.origin

        # These are for eagerly caching webservers
        yield 'Pragma', 'no-cache'
        yield 'Cache-Control', 'no-cache'

        # Request protocols and extensions, these are later checked with the
        # actual supported values from the server's response
        if self.wsock.protocols:
            yield 'Sec-WebSocket-Protocol', ', '.join(self.wsock.protocols)

        if self.wsock.extensions:
            values = [format_param_hdr(e.name, e.request)
                      for e in self.wsock.extensions]
            yield 'Sec-WebSocket-Extensions', ', '.join(values)

    def http_auth_basic_headers(self, **kwargs):
        u, p = self.wsock.auth
        u = u.encode('utf-8')
        p = p.encode('utf-8')
        yield 'Authorization', 'Basic ' + b64encode(u + ':' + p)

    def http_auth_digest_headers(self, **kwargs):
        username, password = self.wsock.auth
        yield 'Authorization', build_authorization_request(
                                username=username.encode('utf-8'),
                                method='GET',
                                uri=self.wsock.location,
                                nonce_count=0,
                                realm=kwargs['realm'],
                                nonce=kwargs['nonce'],
                                opaque=kwargs['opaque'],
                                password=password.encode('utf-8'))


def split_stripped(value, delim=',', maxsplits=-1):
    return map(str.strip, str(value).split(delim, maxsplits)) if value else []


def parse_param_hdr(hdr):
    if ';' in hdr:
        name, paramstr = split_stripped(hdr, ';', 1)
    else:
        name = hdr
        paramstr = ''

    params = {}

    for param in split_stripped(paramstr):
        if '=' in param:
            key, value = split_stripped(param, '=', 1)

            if value.isdigit():
                value = int(value)
        else:
            key = param
            value = True

        params[key] = value

    return name, params


def format_param_hdr(value, params):
    if not params:
        return value

    def fmt_param((k, v)):
        if v is True:
            return k

        if v is not False and v is not None:
            return k + '=' + str(v)

    strparams = filter(None, map(fmt_param, params.items()))
    return '; '.join([value] + strparams)



import socket
import logging
import time
from traceback import format_exc
from threading import Thread
from ssl import SSLError

from websocket import websocket
from connection import Connection
from errors import HandshakeError


class Server(object):
    """
    Websocket server, manages multiple client connections.

    Example usage:
    >>> import wspy

    >>> class EchoServer(wspy.Server):
    >>>     def onopen(self, client):
    >>>         print 'Client %s connected' % client

    >>>     def onmessage(self, client, message):
    >>>         print 'Received message "%s"' % message.payload
    >>>         client.send(wspy.TextMessage(message.payload))

    >>>     def onclose(self, client, code, reason):
    >>>         print 'Client %s disconnected' % client

    >>> EchoServer(('', 8000)).run()
    """

    def __init__(self, address, loglevel=logging.INFO, ssl_args=None,
                 max_join_time=2.0, backlog_size=32, **kwargs):
        """
        Constructor for a simple web socket server.

        `address` is a (hostname, port) tuple to bind the web socket to.

        `loglevel` values should be imported from the logging module.
        logging.INFO only shows server start/stop messages, logging.DEBUG shows
        clients (dis)connecting and messages being sent/received.

        `protocols` and `extensions` are passed directly to the websocket
        constructor.

        `ssl_args` is a dictionary with arguments for `websocket.enable_ssl`
        (and thus to ssl.wrap_socket).  If omitted, the server is not
        SSL-enabled. If specified, at least the dictionary keys "keyfile" and
        "certfile" must be present because these are required arguments for
        `websocket.enable_ssl` for a server socket.

        `max_join_time` is the maximum time (in seconds) to wait for client
        responses after sending CLOSE frames, it defaults to 2 seconds.

        `backlog_size` is directly passed to `websocket.listen`.
        """
        logging.basicConfig(level=loglevel,
                format='%(asctime)s: %(levelname)s: %(message)s',
                datefmt='%H:%M:%S')

        scheme = 'wss' if ssl_args else 'ws'
        hostname, port = address
        logging.info('Starting server at %s://%s:%d', scheme, hostname, port)

        self.sock = websocket(**kwargs)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        if ssl_args:
            self.sock.enable_ssl(server_side=True, **ssl_args)

        self.sock.bind(address)
        self.sock.listen(backlog_size)

        self.max_join_time = max_join_time

    def run(self):
        self.clients = []
        self.client_threads = []

        while True:
            try:
                sock, address = self.sock.accept()

                client = Client(self, sock)
                self.clients.append(client)
                logging.debug('Registered client %s', client)

                thread = Thread(target=client.receive_forever)
                thread.daemon = True
                thread.start()
                self.client_threads.append(thread)
            except SSLError as e:
                logging.error('SSL error: %s', e)
            except HandshakeError as e:
                logging.error('Invalid request: %s', e.message)
            except KeyboardInterrupt:
                logging.info('Received interrupt, stopping server...')
                break
            except Exception as e:
                logging.error(format_exc(e))

        self.quit_gracefully()

    def quit_gracefully(self):
        # Send a CLOSE frame so that the client connection will receive a
        # response CLOSE frame
        for client in self.clients:
            client.send_close_frame()

        # Wait for the CLOSE frames to be received. Wait for all threads in one
        # loop instead of joining separately, so that timeouts are not
        # propagated
        start_time = time.time()

        while time.time() - start_time <= self.max_join_time \
                and any(t.is_alive() for t in self.client_threads):
            time.sleep(0.050)

        # Close remaining sockets, this will trigger a socket.error in the
        # receive_forever() thread, causing the Connection.onclose() handler to
        # be invoked
        for client in self.clients:
            try:
                client.sock.close()
            except socket.error:
                pass

        # Wait for the onclose() handlers to finish
        for thread in self.client_threads:
            thread.join()

    def remove_client(self, client, code, reason):
        self.clients.remove(client)
        self.onclose(client, code, reason)

    def onopen(self, client):
        return NotImplemented

    def onmessage(self, client, message):
        return NotImplemented

    def onping(self, client, payload):
        return NotImplemented

    def onpong(self, client, payload):
        return NotImplemented

    def onclose(self, client, code, reason):
        return NotImplemented

    def onerror(self, client, e):
        return NotImplemented


class Client(Connection):
    def __init__(self, server, sock):
        self.server = server
        super(Client, self).__init__(sock)

    def __str__(self):
        try:
            return '<Client at %s:%d>' % self.sock.getpeername()
        except socket.error:
            return '<Client on closed socket>'

    def send(self, message, fragment_size=None, mask=False):
        logging.debug('Sending %s to %s', message, self)
        Connection.send(self, message, fragment_size=fragment_size, mask=mask)

    def onopen(self):
        logging.debug('Opened socket to %s', self)
        self.server.onopen(self)

    def onmessage(self, message):
        logging.debug('Received %s from %s', message, self)
        self.server.onmessage(self, message)

    def onping(self, payload):
        logging.debug('Sent ping "%s" to %s', payload, self)
        self.server.onping(self, payload)

    def onpong(self, payload):
        logging.debug('Received pong "%s" from %s', payload, self)
        self.server.onpong(self, payload)

    def onclose(self, code, reason):
        msg = 'Closed socket to %s' % self

        if code is not None:
            msg += ': [%d] %s' % (code, reason)

        logging.debug(msg)
        self.server.remove_client(self, code, reason)

    def onerror(self, e):
        logging.error(format_exc(e))
        self.server.onerror(self, e)


if __name__ == '__main__':
    import sys
    port = int(sys.argv[1]) if len(sys.argv) > 1 else 8000
    Server(('', port), loglevel=logging.DEBUG).run()



from websocket import websocket
from server import Server
from frame import Frame, ControlFrame, OPCODE_CONTINUATION, OPCODE_TEXT, \
        OPCODE_BINARY, OPCODE_CLOSE, OPCODE_PING, OPCODE_PONG, CLOSE_NORMAL, \
        CLOSE_GOING_AWAY, CLOSE_PROTOCOL_ERROR, CLOSE_NOACCEPT_DTYPE, \
        CLOSE_INVALID_DATA, CLOSE_POLICY, CLOSE_MESSAGE_TOOBIG, \
        CLOSE_MISSING_EXTENSIONS, CLOSE_UNABLE, read_frame, pop_frame, \
        contains_frame
from connection import Connection
from message import Message, TextMessage, BinaryMessage
from errors import SocketClosed, HandshakeError, PingError, SSLError
from extension import Extension
from deflate_frame import DeflateFrame
from deflate_message import DeflateMessage
from async import AsyncConnection, AsyncServer



from frame import Frame, OPCODE_TEXT, OPCODE_BINARY


__all__ = ['Message', 'TextMessage', 'BinaryMessage']


class Message(object):
    def __init__(self, opcode, payload):
        self.opcode = opcode
        self.payload = payload

    def frame(self, mask=False):
        return Frame(self.opcode, self.payload, mask=mask)

    def fragment(self, fragment_size, mask=False):
        return self.frame().fragment(fragment_size, mask)

    def __str__(self):
        return '<%s opcode=0x%X size=%d>' \
               % (self.__class__.__name__, self.opcode, len(self.payload))


class TextMessage(Message):
    def __init__(self, payload):
        super(TextMessage, self).__init__(OPCODE_TEXT, unicode(payload))

    def frame(self, mask=False):
        return Frame(self.opcode, self.payload.encode('utf-8'), mask=mask)

    def __str__(self):
        if len(self.payload) > 30:
            return '<TextMessage "%s"... size=%d>' \
                    % (self.payload[:30], len(self.payload))

        return '<TextMessage "%s" size=%d>' % (self.payload, len(self.payload))


class BinaryMessage(Message):
    def __init__(self, payload):
        super(BinaryMessage, self).__init__(OPCODE_BINARY, bytearray(payload))


def create_message(opcode, payload):
    if opcode == OPCODE_TEXT:
        return TextMessage(payload.decode('utf-8'))

    if opcode == OPCODE_BINARY:
        return BinaryMessage(payload)

    return Message(opcode, payload)



class SocketClosed(Exception):
    def __init__(self, initialized):
        self.initialized = initialized

    @property
    def message(self):
        s = 'socket closed'

        if self.initialized:
            s += ' (initialized)'

        return s


class HandshakeError(Exception):
    pass


class PingError(Exception):
    pass


class SSLError(Exception):
    pass



from extension import Extension


class Multiplex(Extension):
    """
    This is an implementation of the "mux" extension, as defined by
    http://tools.ietf.org/html/draft-ietf-hybi-websocket-multiplexing-11.

    Supported parameters are:
    - quota: TODO
    """

    name = 'mux'
    rsv1 = True  # FIXME
    rsv2 = True  # FIXME
    rsv3 = True  # FIXME
    defaults = {'quota': None}

    def __init__(self, defaults={}, request={}):
        Extension.__init__(self, defaults, request)

        # TODO: check "quota" value

    class Hook(Extension.Hook):
        def send(self, frame):
            raise NotImplementedError  # TODO

        def recv(self, frame):
            raise NotImplementedError  # TODO



import socket

from frame import ControlFrame, OPCODE_CLOSE, OPCODE_PING, OPCODE_PONG, \
                  OPCODE_CONTINUATION, create_close_frame
from message import create_message
from errors import SocketClosed, PingError


class Connection(object):
    """
    A `Connection` uses a `websocket` instance to send and receive (optionally
    fragmented) messages, which are `Message` instances. Control frames are
    handled automatically in the way specified by RFC 6455.

    To use the `Connection` class, it should be extended and the extending
    class should implement the on*() event handlers.

    Example of an echo server (sends back what it receives):
    >>> import wspy

    >>> class EchoConnection(wspy.Connection):
    >>>     def onopen(self):
    >>>         print 'Connection opened at %s:%d' % self.sock.getpeername()

    >>>     def onmessage(self, message):
    >>>         print 'Received message "%s"' % message.payload
    >>>         self.send(wspy.TextMessage(message.payload))

    >>>     def onclose(self, code, reason):
    >>>         print 'Connection closed'

    >>> server = wspy.websocket()
    >>> server.bind(('', 8000))
    >>> server.listen()

    >>> while True:
    >>>     client, addr = server.accept()
    >>>     EchoConnection(client).receive_forever()
    """
    def __init__(self, sock):
        """
        `sock` is a websocket instance which has completed its handshake.
        """
        self.sock = sock

        self.close_frame_sent = False
        self.close_frame_received = False
        self.ping_sent = False
        self.ping_payload = None

        self.hooks_send = []
        self.hooks_recv = []

        self.onopen()

    def message_to_frames(self, message, fragment_size=None, mask=False):
        frame = self.sock.apply_send_hooks(message.frame(mask=mask), True)

        if fragment_size is None:
            yield frame
        else:
            for fragment in frame.fragment(fragment_size):
                yield fragment

    def send(self, message, fragment_size=None, mask=False):
        """
        Send a message. If `fragment_size` is specified, the message is
        fragmented into multiple frames whose payload size does not extend
        `fragment_size`.
        """
        for frame in self.message_to_frames(message, fragment_size, mask):
            self.send_frame(frame)

    def send_frame(self, frame, callback=None):
        self.sock.send(frame)

        if callback:
            callback()

    def recv(self):
        """
        Receive a message. A message may consist of multiple (ordered) data
        frames. A control frame may be delivered at any time, also when
        expecting the next continuation frame of a fragmented message. These
        control frames are handled immediately by handle_control_frame().
        """
        fragments = []

        while not len(fragments) or not fragments[-1].final:
            frame = self.sock.recv()

            if isinstance(frame, ControlFrame):
                self.handle_control_frame(frame)
            elif len(fragments) > 0 and frame.opcode != OPCODE_CONTINUATION:
                raise ValueError('expected continuation/control frame, got %s '
                                 'instead' % frame)
            else:
                fragments.append(frame)

        return self.concat_fragments(fragments)

    def concat_fragments(self, fragments):
        frame = fragments[0]

        for f in fragments[1:]:
            frame.payload += f.payload

        frame.final = True
        frame = self.sock.apply_recv_hooks(frame, True)
        return create_message(frame.opcode, frame.payload)

    def handle_control_frame(self, frame):
        """
        Handle a control frame as defined by RFC 6455.
        """
        if frame.opcode == OPCODE_CLOSE:
            self.close_frame_received = True
            code, reason = frame.unpack_close()

            if self.close_frame_sent:
                self.onclose(code, reason)
                self.sock.close()
                raise SocketClosed(True)
            else:
                self.close_params = (code, reason)
                self.send_close_frame(code, reason)

        elif frame.opcode == OPCODE_PING:
            # Respond with a pong message with identical payload
            self.send_frame(ControlFrame(OPCODE_PONG, frame.payload))

        elif frame.opcode == OPCODE_PONG:
            # Assert that the PONG payload is identical to that of the PING
            if not self.ping_sent:
                raise PingError('received PONG while no PING was sent')

            self.ping_sent = False

            if frame.payload != self.ping_payload:
                raise PingError('received PONG with invalid payload')

            self.ping_payload = None
            self.onpong(frame.payload)

    def receive_forever(self):
        """
        Receive and handle messages in an endless loop. A message may consist
        of multiple data frames, but this is not visible for onmessage().
        Control messages (or control frames) are handled automatically.
        """
        while True:
            try:
                self.onmessage(self.recv())
            except (KeyboardInterrupt, SystemExit, SocketClosed):
                break
            except Exception as e:
                self.onerror(e)
                self.onclose(None, 'error: %s' % e)

                try:
                    self.sock.close()
                except socket.error:
                    pass

                raise e

    def send_ping(self, payload=''):
        """
        Send a PING control frame with an optional payload.
        """
        self.send_frame(ControlFrame(OPCODE_PING, payload),
                        lambda: self.onping(payload))
        self.ping_payload = payload
        self.ping_sent = True

    def send_close_frame(self, code, reason):
        self.send_frame(create_close_frame(code, reason))
        self.close_frame_sent = True
        self.shutdown_write()

    def shutdown_write(self):
        if self.close_frame_received:
            self.onclose(*self.close_params)
            self.sock.close()
            raise SocketClosed(False)
        else:
            self.sock.shutdown(socket.SHUT_WR)

    def close(self, code=None, reason=''):
        """
        Close the socket by sending a CLOSE frame and waiting for a response
        close message, unless such a message has already been received earlier
        (prior to calling this function, for example). The onclose() handler is
        called after the response has been received, but before the socket is
        actually closed.
        """
        self.send_close_frame(code, reason)

        frame = self.sock.recv()

        if frame.opcode != OPCODE_CLOSE:
            raise ValueError('expected CLOSE frame, got %s' % frame)

        self.handle_control_frame(frame)

    def onopen(self):
        """
        Called after the connection is initialized.
        """
        return NotImplemented

    def onmessage(self, message):
        """
        Called when a message is received. `message` is a Message object, which
        can be constructed from a single frame or multiple fragmented frames.
        """
        return NotImplemented

    def onping(self, payload):
        """
        Called after a PING control frame has been sent. This handler could be
        used to start a timeout handler for a PONG frame that is not received
        in time.
        """
        return NotImplemented

    def onpong(self, payload):
        """
        Called when a PONG control frame is received.
        """
        return NotImplemented

    def onclose(self, code, reason):
        """
        Called when the socket is closed by either end point.
        """
        return NotImplemented

    def onerror(self, e):
        """
        Handle a raised exception.
        """
        return NotImplemented



class Extension(object):
    name = ''
    rsv1 = False
    rsv2 = False
    rsv3 = False
    opcodes = ()
    before_fragmentation = False
    defaults = {}

    def __init__(self, **kwargs):
        for param in kwargs.iterkeys():
            if param not in self.defaults:
                raise KeyError('unrecognized parameter "%s"' % param)

        # Copy dict first to avoid duplicate references to the same object
        self.defaults = dict(self.__class__.defaults)
        self.defaults.update(kwargs)

    def __str__(self):
        return '<Extension "%s" defaults=%s request=%s>' \
               % (self.name, self.defaults, self.request)

    @property
    def names(self):
        return (self.name,) if self.name else ()

    def is_supported(self, name, other_instances):
        return name in self.names and not any(self.conflicts(other.extension)
                                              for other in other_instances)

    def conflicts(self, ext):
        """
        Check if the extension conflicts with an already accepted extension.
        This may be the case when the two extensions use the same reserved
        bits, or have the same name (when the same extension is negotiated
        multiple times with different parameters).
        """
        return ext.rsv1 and self.rsv1 \
            or ext.rsv2 and self.rsv2 \
            or ext.rsv3 and self.rsv3 \
            or set(ext.names) & set(self.names) \
            or set(ext.opcodes) & set(self.opcodes)

    def negotiate(self, name, params):
        """
        Same as `negotiate_safe`, but instead returns an iterator of (param,
        value) tuples and raises an exception on error.
        """
        raise NotImplementedError

    def negotiate_safe(self, name, params):
        """
        `name` and `params` are sent in the HTTP request by the client. Check
        if the extension name is supported by this extension, and validate the
        parameters. Returns a dict with accepted parameters, or None if not
        accepted.
        """
        for param in params.iterkeys():
            if param not in self.defaults:
                return

        try:
            return dict(self.negotiate(name, params))
        except (KeyError, ValueError, AssertionError):
            pass

    class Instance:
        def __init__(self, extension, name, params):
            self.extension = extension
            self.name = name
            self.params = params

            for param, value in extension.defaults.iteritems():
                setattr(self, param, value)

            for param, value in params.iteritems():
                setattr(self, param, value)

            self.init()

        def init(self):
            return NotImplemented

        def handle_send(self, frame):
            if self.extension.before_fragmentation:
                assert not frame.is_fragmented()

            replacement = self.onsend(frame)
            return frame if replacement is None else replacement

        def handle_recv(self, frame):
            if self.extension.before_fragmentation:
                assert not frame.is_fragmented()

            replacement = self.onrecv(frame)
            return frame if replacement is None else replacement

        def onsend(self, frame):
            raise NotImplementedError

        def onrecv(self, frame):
            raise NotImplementedError



import struct
import socket
from os import urandom
from string import printable


OPCODE_CONTINUATION = 0x0
OPCODE_TEXT = 0x1
OPCODE_BINARY = 0x2
OPCODE_CLOSE = 0x8
OPCODE_PING = 0x9
OPCODE_PONG = 0xA

CLOSE_NORMAL = 1000
CLOSE_GOING_AWAY = 1001
CLOSE_PROTOCOL_ERROR = 1002
CLOSE_NOACCEPT_DTYPE = 1003
CLOSE_INVALID_DATA = 1007
CLOSE_POLICY = 1008
CLOSE_MESSAGE_TOOBIG = 1009
CLOSE_MISSING_EXTENSIONS = 1010
CLOSE_UNABLE = 1011

line_printable = [c for c in printable if c not in '\r\n\x0b\x0c']


def printstr(s):
    return ''.join(c if c in line_printable else '.' for c in str(s))


class Frame(object):
    """
    A Frame instance represents a web socket data frame as defined in RFC 6455.
    To encoding a frame for sending it over a socket, use Frame.pack(). To
    receive and decode a frame from a socket, use receive_frame().
    """
    def __init__(self, opcode, payload, masking_key='', mask=False, final=True,
            rsv1=False, rsv2=False, rsv3=False):
        """
        Create a new frame.

        `opcode` is one of the constants as defined above.

        `payload` is a string of bytes containing the data sendt in the frame.

        `masking_key` is an optional custom key to use for masking, or `mask`
        can be used instead to let this constructor generate a random masking
        key.

        `final` is a boolean indicating whether this frame is the last in a
        chain of fragments.

        `rsv1`, `rsv2` and `rsv3` are booleans indicating bit values for RSV1,
        RVS2 and RSV3, which are only non-zero if defined so by extensions.
        """
        if mask:
            masking_key = urandom(4)

        if len(masking_key) not in (0, 4):
            raise ValueError('invalid masking key "%s"' % masking_key)

        self.final = final
        self.rsv1 = rsv1
        self.rsv2 = rsv2
        self.rsv3 = rsv3
        self.opcode = opcode
        self.masking_key = masking_key
        self.payload = payload

    def pack(self):
        """
        Pack the frame into a string according to the following scheme:

        +-+-+-+-+-------+-+-------------+-------------------------------+
        |F|R|R|R| opcode|M| Payload len |    Extended payload length    |
        |I|S|S|S|  (4)  |A|     (7)     |             (16/64)           |
        |N|V|V|V|       |S|             |   (if payload len==126/127)   |
        | |1|2|3|       |K|             |                               |
        +-+-+-+-+-------+-+-------------+ - - - - - - - - - - - - - - - +
        |     Extended payload length continued, if payload len == 127  |
        + - - - - - - - - - - - - - - - +-------------------------------+
        |                               |Masking-key, if MASK set to 1  |
        +-------------------------------+-------------------------------+
        | Masking-key (continued)       |          Payload Data         |
        +-------------------------------- - - - - - - - - - - - - - - - +
        :                     Payload Data continued ...                :
        + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        |                     Payload Data continued ...                |
        +---------------------------------------------------------------+
        """
        header = struct.pack('!B', (self.final << 7) | (self.rsv1 << 6)
                                   | (self.rsv2 << 5) | (self.rsv3 << 4)
                                   | (self.opcode & 0xf))
        mask = bool(self.masking_key) << 7
        payload_len = len(self.payload)

        if payload_len <= 125:
            header += struct.pack('!B', mask | payload_len)
        elif payload_len < (1 << 16):
            header += struct.pack('!BH', mask | 126, payload_len)
        elif payload_len < (1 << 63):
            header += struct.pack('!BQ', mask | 127, payload_len)
        else:
            # FIXME: RFC 6455 defines an action for this...
            raise Exception('the payload length is too damn high!')

        if mask:
            return header + self.masking_key + self.mask_payload()

        return header + self.payload

    def mask_payload(self):
        return mask(self.masking_key, self.payload)

    def fragment(self, fragment_size, mask=False):
        """
        Fragment the frame into a chain of fragment frames:
        - An initial frame with non-zero opcode
        - Zero or more frames with opcode = 0 and final = False
        - A final frame with opcode = 0 and final = True

        The first and last frame may be the same frame, having a non-zero
        opcode and final = True. Thus, this function returns a list containing
        at least a single frame.

        `fragment_size` indicates the maximum payload size of each fragment.
        The payload of the original frame is split into one or more parts, and
        each part is converted to a Frame instance.

        `mask` is a boolean (default False) indicating whether the payloads
        should be masked. If True, each frame is assigned a randomly generated
        masking key.
        """
        frames = []

        for start in xrange(0, len(self.payload), fragment_size):
            payload = self.payload[start:start + fragment_size]
            frames.append(Frame(OPCODE_CONTINUATION, payload, mask=mask,
                                final=False))

        frames[0].opcode = self.opcode
        frames[-1].final = True

        return frames

    def is_fragmented(self):
        return not self.final or self.opcode == OPCODE_CONTINUATION

    def __str__(self):
        s = '<%s opcode=0x%X len=%d' \
            % (self.__class__.__name__, self.opcode, len(self.payload))

        if self.masking_key:
            s += ' masking_key=%4s' % printstr(self.masking_key)

        max_pl_disp = 30
        pl = printstr(self.payload)[:max_pl_disp]

        if len(self.payload) > max_pl_disp:
            pl += '...'

        s += ' payload=%s' % pl

        if self.rsv1:
            s += ' rsv1'

        if self.rsv2:
            s += ' rsv2'

        if self.rsv3:
            s += ' rsv3'

        return s + '>'


class ControlFrame(Frame):
    """
    A control frame is a frame with an opcode OPCODE_CLOSE, OPCODE_PING or
    OPCODE_PONG. These frames must be handled as defined by RFC 6455, and
    """
    def fragment(self, fragment_size, mask=False):
        """
        Control frames must not be fragmented.
        """
        raise TypeError('control frames must not be fragmented')

    def pack(self):
        """
        Same as Frame.pack(), but asserts that the payload size does not exceed
        125 bytes.
        """
        if len(self.payload) > 125:
            raise ValueError('control frames must not be larger than 125 '
                             'bytes')

        return Frame.pack(self)

    def unpack_close(self):
        """
        Unpack a close message into a status code and a reason. If no payload
        is given, the code is None and the reason is an empty string.
        """
        if self.payload:
            code = struct.unpack('!H', str(self.payload[:2]))[0]
            reason = str(self.payload[2:])
        else:
            code = None
            reason = ''

        return code, reason


def decode_frame(reader):
    b1, b2 = struct.unpack('!BB', reader.readn(2))

    final = bool(b1 & 0x80)
    rsv1 = bool(b1 & 0x40)
    rsv2 = bool(b1 & 0x20)
    rsv3 = bool(b1 & 0x10)
    opcode = b1 & 0x0F

    masked = bool(b2 & 0x80)
    payload_len = b2 & 0x7F

    if payload_len == 126:
        payload_len = struct.unpack('!H', reader.readn(2))
    elif payload_len == 127:
        payload_len = struct.unpack('!Q', reader.readn(8))

    if masked:
        masking_key = reader.readn(4)
        payload = mask(masking_key, reader.readn(payload_len))
    else:
        masking_key = ''
        payload = reader.readn(payload_len)

    # Control frames have most significant bit 1
    cls = ControlFrame if opcode & 0x8 else Frame

    return cls(opcode, payload, masking_key=masking_key, final=final,
               rsv1=rsv1, rsv2=rsv2, rsv3=rsv3)


def receive_frame(sock):
    return decode_frame(SocketReader(sock))


def read_frame(data):
    reader = BufferReader(data)
    frame = decode_frame(reader)
    return frame, reader.offset


def pop_frame(data):
    frame, size = read_frame(data)
    return frame, data[size:]


class BufferReader(object):
    def __init__(self, data):
        self.data = data
        self.offset = 0

    def readn(self, n):
        assert len(self.data) - self.offset >= n
        self.offset += n
        return self.data[self.offset - n:self.offset]


class SocketReader(object):
    def __init__(self, sock):
        self.sock = sock

    def readn(self, n):
        """
        Keep receiving data until exactly `n` bytes have been read.
        """
        data = ''

        while len(data) < n:
            received = self.sock.recv(n - len(data))

            if not len(received):
                raise socket.error('no data read from socket')

            data += received

        return data


def contains_frame(data):
    """
    Read the frame length from the start of `data` and check if the data is
    long enough to contain the entire frame.
    """
    if len(data) < 2:
        return False

    b2 = struct.unpack('!B', data[1])[0]
    payload_len = b2 & 0x7F
    payload_start = 2

    if payload_len == 126:
        if len(data) > 4:
            payload_len = struct.unpack('!H', data[2:4])

        payload_start = 4
    elif payload_len == 127:
        if len(data) > 12:
            payload_len = struct.unpack('!Q', data[4:12])

        payload_start = 12

    return len(data) >= payload_len + payload_start


def mask(key, original):
    """
    Mask an octet string using the given masking key.
    The following masking algorithm is used, as defined in RFC 6455:

    for each octet:
        j = i MOD 4
        transformed-octet-i = original-octet-i XOR masking-key-octet-j
    """
    if len(key) != 4:
        raise ValueError('invalid masking key "%s"' % key)

    key = map(ord, key)
    masked = bytearray(original)

    for i in xrange(len(masked)):
        masked[i] ^= key[i % 4]

    return masked


def create_close_frame(code, reason):
    payload = '' if code is None else struct.pack('!H', code) + reason
    return ControlFrame(OPCODE_CLOSE, payload)



#!/usr/bin/env python
from setuptools import setup, find_packages


setup(
    name='wspy',
    version='0.9.1',
    description='A standalone implementation of websockets (RFC 6455).',
    author='Taddeus Kroes',
    author_email='taddeus@kompiler.org',
    url='https://github.com/taddeus/wspy',
    package_dir={'wspy': '.'},
    packages=['wspy'],
    license='3-clause BSD License'
)



'''
Copyright (c) 2009, Akoha, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of python-digest nor the names of its contributors may be
   used to endorse or promote products derived from this software without
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This code was downloaded from https://github.com/dimagi/python-digest to avoid
having to download python-digest as a dependency for wspy.
'''

try:
    import hashlib as md5
except ImportError: # Python <2.5
    import md5

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

import random
import types
import urllib
import urlparse
import logging

# Make sure a NullHandler is available
# This was added in Python 2.7/3.2
try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

_REQUIRED_DIGEST_RESPONSE_PARTS = ['username', 'realm', 'nonce', 'uri', 'response', 'algorithm',
                  'opaque', 'qop', 'nc', 'cnonce']
_REQUIRED_DIGEST_CHALLENGE_PARTS = ['realm', 'nonce', 'stale', 'algorithm',
                             'opaque', 'qop']

l = logging.getLogger(__name__)
l.addHandler(NullHandler())

_LWS=[chr(9), ' ', '\r', '\n']
_ILLEGAL_TOKEN_CHARACTERS = (
    [chr(n) for n in range(0-31)] + # control characters
    [chr(127)] + # DEL
    ['(',')','<','>','@',',',';',':','\\','"','/','[',']','?','=','{','}',' '] +
    [chr(9)]) # horizontal tab

class State(object):
    def character(self, c):
        return self.consume(c)

    def close(self):
        return self.eof()

    def eof(self):
        raise ValueError('EOF not permitted in this state.')

    '''
    Return False to keep the current state, or True to pop it
    '''
    def consume(c):
        raise Exception('Unimplemented')

class ParentState(State):
    def __init__(self):
        super(State, self).__init__()
        self.child = None

    def close(self):
        if self.child:
            return self.handle_child_return(self.child.close())
        else:
            return self.eof()

    def push_child(self, child, c=None):
        self.child = child
        if c is not None:
            return self.send_to_child(c)
        else:
            return False

    def send_to_child(self, c):
        return self.handle_child_return(self.child.character(c))

    def handle_child_return(self, returned_value):
        if returned_value:
            child = self.child
            self.child = None
            return self.child_complete(child)
        return False

    '''
    Return False to keep the current state, or True to pop it.
    '''
    def child_complete(self, child):
        return False

    def character(self, c):
        if self.child:
            return self.send_to_child(c)
        else:
            return self.consume(c)

    def consume(self, c):
        return False


class EscapedCharacterState(State):
    def __init__(self, io):
        super(EscapedCharacterState, self).__init__()
        self.io = io

    def consume(self, c):
        self.io.write(c)
        return True

class KeyTrailingWhitespaceState(State):
    def consume(self, c):
        if c in _LWS:
            return False
        elif c == '=':
            return True
        else:
            raise ValueError("Expected whitespace or '='")

class ValueLeadingWhitespaceState(ParentState):
    def __init__(self, io):
        super(ValueLeadingWhitespaceState, self).__init__()
        self.io = io

    def consume(self, c):
        if c in _LWS:
            return False
        elif c == '"':
            return self.push_child(QuotedValueState(self.io))
        elif c in _ILLEGAL_TOKEN_CHARACTERS:
            raise ValueError('The character %r is not a legal token character' % c)
        else:
            self.io.write(c)
            return self.push_child(UnquotedValueState(self.io))

    def child_complete(self, child):
        return True

class ValueTrailingWhitespaceState(State):
    def eof(self):
        return True

    def consume(self, c):
        if c in _LWS:
            return False
        elif c == ',':
            return True
        else:
            raise ValueError("Expected whitespace, ',', or EOF")

class BaseQuotedState(ParentState):
    def __init__(self, io):
        super(BaseQuotedState, self).__init__()
        self.key_io = io

    def consume(self, c):
        if c == '\\':
            return self.push_child(EscapedCharacterState(self.key_io))
        elif c == '"':
            return self.push_child(self.TrailingWhitespaceState())
        else:
            self.key_io.write(c)
            return False

    def child_complete(self, child):
        if type(child) == EscapedCharacterState:
            return False
        elif type(child) == self.TrailingWhitespaceState:
            return True

class BaseUnquotedState(ParentState):
    def __init__(self, io):
        super(BaseUnquotedState, self).__init__()
        self.io = io

    def consume(self, c):
        if c == self.terminating_character:
            return True
        elif c in _LWS:
            return self.push_child(self.TrailingWhitespaceState())
        elif c in _ILLEGAL_TOKEN_CHARACTERS:
            raise ValueError('The character %r is not a legal token character' % c)
        else:
            self.io.write(c)
            return False

    def child_complete(self, child):
        # type(child) == self.TrailingWhitespaceState
        return True

class QuotedKeyState(BaseQuotedState):
    TrailingWhitespaceState = KeyTrailingWhitespaceState

class QuotedValueState(BaseQuotedState):
    TrailingWhitespaceState = ValueTrailingWhitespaceState

class UnquotedKeyState(BaseUnquotedState):
    TrailingWhitespaceState = KeyTrailingWhitespaceState
    terminating_character = '='

class UnquotedValueState(BaseUnquotedState):
    TrailingWhitespaceState = ValueTrailingWhitespaceState
    terminating_character  = ','

    def eof(self):
        return True

class NewPartState(ParentState):
    def __init__(self, parts):
        super(NewPartState, self).__init__()
        self.parts = parts
        self.key_io = StringIO()
        self.value_io = StringIO()

    def consume(self, c):
        if c in _LWS:
            return False
        elif c == '"':
            return self.push_child(QuotedKeyState(self.key_io))
        elif c in _ILLEGAL_TOKEN_CHARACTERS:
            raise ValueError('The character %r is not a legal token character' % c)
        else:
            self.key_io.write(c)
            return self.push_child(UnquotedKeyState(self.key_io))

    def child_complete(self, child):
        if type(child) in [QuotedKeyState, UnquotedKeyState]:
            return self.push_child(ValueLeadingWhitespaceState(self.value_io))
        else:
            self.parts[self.key_io.getvalue()] = self.value_io.getvalue()
            return True

class FoundationState(ParentState):
    def __init__(self, defaults):
        super(FoundationState, self).__init__()
        self.parts = defaults.copy()

    def result(self):
        return self.parts

    def consume(self, c):
        return self.push_child(NewPartState(self.parts), c)

def parse_parts(parts_string, defaults={}):
    state_machine = FoundationState(defaults)
    index = 0
    try:
        for c in parts_string:
            state_machine.character(c)
            index += 1
        state_machine.close()
        return state_machine.result()
    except ValueError, e:
        annotated_parts_string = "%s[%s]%s" % (parts_string[0:index],
                                               index < len(parts_string) and parts_string[index] or '',
                                               index + 1 < len(parts_string) and parts_string[index+1:] or '')
        l.exception("Failed to parse the Digest string "
                    "(offending character is in []): %r" % annotated_parts_string)
        return None

def format_parts(**kwargs):
    return ", ".join(['%s="%s"' % (k,v.encode('utf-8')) for (k,v) in kwargs.items()])

def validate_uri(digest_uri, request_path):
    digest_url_components = urlparse.urlparse(digest_uri)
    return urllib.unquote(digest_url_components[2]) == request_path

def validate_nonce(nonce, secret):
    '''
    Is the nonce one that was generated by this library using the provided secret?
    '''
    nonce_components = nonce.split(':', 2)
    if not len(nonce_components) == 3:
        return False
    timestamp = nonce_components[0]
    salt = nonce_components[1]
    nonce_signature = nonce_components[2]

    calculated_nonce = calculate_nonce(timestamp, secret, salt)

    if not nonce == calculated_nonce:
        return False

    return True

def calculate_partial_digest(username, realm, password):
    '''
    Calculate a partial digest that may be stored and used to authenticate future
    HTTP Digest sessions.
    '''
    return md5.md5("%s:%s:%s" % (username.encode('utf-8'), realm, password.encode('utf-8'))).hexdigest()

def build_digest_challenge(timestamp, secret, realm, opaque, stale):
    '''
    Builds a Digest challenge that may be sent as the value of the 'WWW-Authenticate' header
    in a 401 or 403 response.

    'opaque' may be any value - it will be returned by the client.

    'timestamp' will be incorporated and signed in the nonce - it may be retrieved from the
    client's authentication request using get_nonce_timestamp()
    '''
    nonce = calculate_nonce(timestamp, secret)

    return 'Digest %s' % format_parts(realm=realm, qop='auth', nonce=nonce,
                                      opaque=opaque, algorithm='MD5',
                                      stale=stale and 'true' or 'false')

def calculate_request_digest(method, partial_digest, digest_response=None,
                             uri=None, nonce=None, nonce_count=None, client_nonce=None):
    '''
    Calculates a value for the 'response' value of the client authentication request.
    Requires the 'partial_digest' calculated from the realm, username, and password.

    Either call it with a digest_response to use the values from an authentication request,
    or pass the individual parameters (i.e. to generate an authentication request).
    '''
    if digest_response:
        if uri or nonce or nonce_count or client_nonce:
            raise Exception("Both digest_response and one or more "
                            "individual parameters were sent.")
        uri = digest_response.uri
        nonce = digest_response.nonce
        nonce_count = digest_response.nc
        client_nonce=digest_response.cnonce
    elif not (uri and nonce and (nonce_count != None) and client_nonce):
        raise Exception("Neither digest_response nor all individual parameters were sent.")

    ha2 = md5.md5("%s:%s" % (method, uri)).hexdigest()
    data = "%s:%s:%s:%s:%s" % (nonce, "%08x" % nonce_count, client_nonce, 'auth', ha2)
    kd = md5.md5("%s:%s" % (partial_digest, data)).hexdigest()
    return kd

def get_nonce_timestamp(nonce):
    '''
    Extract the timestamp from a Nonce. To be sure the timestamp was generated by this site,
    make sure you validate the nonce using validate_nonce().
    '''
    components = nonce.split(':',2)
    if not len(components) == 3:
        return None

    try:
        return float(components[0])
    except ValueError:
        return None

def calculate_nonce(timestamp, secret, salt=None):
    '''
    Generate a nonce using the provided timestamp, secret, and salt. If the salt is not provided,
    (and one should only be provided when validating a nonce) one will be generated randomly
    in order to ensure that two simultaneous requests do not generate identical nonces.
    '''
    if not salt:
        salt = ''.join([random.choice('0123456789ABCDEF') for x in range(4)])
    return "%s:%s:%s" % (timestamp, salt,
                         md5.md5("%s:%s:%s" % (timestamp, salt, secret)).hexdigest())

def build_authorization_request(username, method, uri, nonce_count, digest_challenge=None,
                                realm=None, nonce=None, opaque=None, password=None,
                                request_digest=None, client_nonce=None):
    '''
    Builds an authorization request that may be sent as the value of the 'Authorization'
    header in an HTTP request.

    Either a digest_challenge object (as returned from parse_digest_challenge) or its required
    component parameters (nonce, realm, opaque) must be provided.

    The nonce_count should be the last used nonce_count plus one.

    Either the password or the request_digest should be provided - if provided, the password
    will be used to generate a request digest. The client_nonce is optional - if not provided,
    a random value will be generated.
    '''
    if not client_nonce:
        client_nonce =  ''.join([random.choice('0123456789ABCDEF') for x in range(32)])

    if digest_challenge and (realm or nonce or opaque):
        raise Exception("Both digest_challenge and one or more of realm, nonce, and opaque"
                        "were sent.")

    if digest_challenge:
        if isinstance(digest_challenge, types.StringType):
            digest_challenge_header = digest_challenge
            digest_challenge = parse_digest_challenge(digest_challenge_header)
            if not digest_challenge:
                raise Exception("The provided digest challenge header could not be parsed: %s" %
                                digest_challenge_header)
        realm = digest_challenge.realm
        nonce = digest_challenge.nonce
        opaque = digest_challenge.opaque
    elif not (realm and nonce and opaque):
        raise Exception("Either digest_challenge or realm, nonce, and opaque must be sent.")

    if password and request_digest:
        raise Exception("Both password and calculated request_digest were sent.")
    elif not request_digest:
        if not password:
            raise Exception("Either password or calculated request_digest must be provided.")

        partial_digest = calculate_partial_digest(username, realm, password)
        request_digest = calculate_request_digest(method, partial_digest, uri=uri, nonce=nonce,
                                                  nonce_count=nonce_count,
                                                  client_nonce=client_nonce)

    return 'Digest %s' % format_parts(username=username, realm=realm, nonce=nonce, uri=uri,
                                      response=request_digest, algorithm='MD5', opaque=opaque,
                                      qop='auth', nc='%08x' % nonce_count, cnonce=client_nonce)

def _check_required_parts(parts, required_parts):
    if parts == None:
        return False

    missing_parts = [part for part in required_parts if not part in parts]
    return len(missing_parts) == 0

def _build_object_from_parts(parts, names):
    obj = type("", (), {})()
    for part_name in names:
        val = parts[part_name]
        if isinstance(val, basestring):
            val = unicode(val, "utf-8")
        setattr(obj, part_name, val)
    return obj

def parse_digest_response(digest_response_string):
    '''
    Parse the parameters of a Digest response. The input is a comma separated list of
    token=(token|quoted-string). See RFCs 2616 and 2617 for details.

    Known issue: this implementation will fail if there are commas embedded in quoted-strings.
    '''

    parts = parse_parts(digest_response_string, defaults={'algorithm': 'MD5'})
    if not _check_required_parts(parts, _REQUIRED_DIGEST_RESPONSE_PARTS):
        return None

    if not parts['nc'] or [c for c in parts['nc'] if not c in '0123456789abcdefABCDEF']:
        return None
    parts['nc'] = int(parts['nc'], 16)

    digest_response = _build_object_from_parts(parts, _REQUIRED_DIGEST_RESPONSE_PARTS)
    if ('MD5', 'auth') != (digest_response.algorithm, digest_response.qop):
        return None

    return digest_response

def is_digest_credential(authorization_header):
    '''
    Determines if the header value is potentially a Digest response sent by a client (i.e.
    if it starts with 'Digest ' (case insensitive).
    '''
    return authorization_header[:7].lower() == 'digest '

def parse_digest_credentials(authorization_header):
    '''
    Parses the value of an 'Authorization' header. Returns an object with properties
    corresponding to each of the recognized parameters in the header.
    '''
    if not is_digest_credential(authorization_header):
        return None

    return parse_digest_response(authorization_header[7:])

def is_digest_challenge(authentication_header):
    '''
    Determines if the header value is potentially a Digest challenge sent by a server (i.e.
    if it starts with 'Digest ' (case insensitive).
    '''
    return authentication_header[:7].lower() == 'digest '

def parse_digest_challenge(authentication_header):
    '''
    Parses the value of a 'WWW-Authenticate' header. Returns an object with properties
    corresponding to each of the recognized parameters in the header.
    '''
    if not is_digest_challenge(authentication_header):
        return None

    parts = parse_parts(authentication_header[7:], defaults={'algorithm': 'MD5',
                                                             'stale': 'false'})
    if not _check_required_parts(parts, _REQUIRED_DIGEST_CHALLENGE_PARTS):
        return None

    parts['stale'] = parts['stale'].lower() == 'true'

    digest_challenge = _build_object_from_parts(parts, _REQUIRED_DIGEST_CHALLENGE_PARTS)
    if ('MD5', 'auth') != (digest_challenge.algorithm, digest_challenge.qop):
        return None

    return digest_challenge




import zlib

from extension import Extension
from deflate_frame import DeflateFrame


class DeflateMessage(Extension):
    """
    Implementation of the "permessage-deflate" extension, as defined by
    http://tools.ietf.org/html/draft-ietf-hybi-permessage-compression-17.

    Note: this implementetion is only eligible for server sockets, client
    sockets must NOT use it.
    """
    name = 'permessage-deflate'
    rsv1 = True
    defaults = {
        'client_max_window_bits': zlib.MAX_WBITS,
        'client_no_context_takeover': False,
        'server_max_window_bits': zlib.MAX_WBITS,
        'server_no_context_takeover': False
    }
    before_fragmentation = True

    compression_threshold = 20  # minimal message payload size for compression

    def negotiate(self, name, params):
        default = self.defaults['client_max_window_bits']

        if 'client_max_window_bits' in params:
            mwb = params['client_max_window_bits']

            if mwb is True:
                if default != zlib.MAX_WBITS:
                    yield 'client_max_window_bits', default
            else:
                mwb = int(mwb)
                assert 8 <= mwb <= zlib.MAX_WBITS
                yield 'client_max_window_bits', min(mwb, default)
        elif default != zlib.MAX_WBITS:
            yield 'client_max_window_bits', default

        if 'client_no_context_takeover' in params:
            assert params['client_no_context_takeover'] is True
            yield 'client_no_context_takeover', True
        elif self.defaults['client_no_context_takeover']:
            yield 'client_no_context_takeover', True

        default = self.defaults['server_max_window_bits']

        if 'server_max_window_bits' in params:
            mwb = int(params['server_max_window_bits'])
            assert 8 <= mwb <= zlib.MAX_WBITS
            yield 'server_max_window_bits', min(mwb, default)
        elif default != zlib.MAX_WBITS:
            yield 'server_max_window_bits', default

        if 'server_no_context_takeover' in params:
            assert params['server_no_context_takeover'] is True
            yield 'server_no_context_takeover', True
        elif self.defaults['server_no_context_takeover']:
            yield 'server_no_context_takeover', True

    class Instance(DeflateFrame.Instance):
        def init(self):
            if not self.server_no_context_takeover:
                self.defl = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                        zlib.DEFLATED, -self.server_max_window_bits)

            if not self.client_no_context_takeover:
                self.dec = zlib.decompressobj(-self.client_max_window_bits)

        def deflate(self, data):
            if self.server_no_context_takeover:
                self.defl = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                        zlib.DEFLATED, -self.server_max_window_bits)

            compressed = self.defl.compress(data)
            compressed += self.defl.flush(zlib.Z_SYNC_FLUSH)
            assert compressed[-4:] == '\x00\x00\xff\xff'
            return compressed[:-4]

        def inflate(self, data):
            data = str(data + '\x00\x00\xff\xff')

            if self.client_no_context_takeover:
                self.dec = zlib.decompressobj(-self.client_max_window_bits)

            return self.dec.decompress(data)



import socket
from select import epoll, EPOLLIN, EPOLLOUT, EPOLLHUP
from traceback import format_exc
import logging

from connection import Connection
from frame import ControlFrame, OPCODE_PING, OPCODE_CONTINUATION, \
                  create_close_frame
from server import Server, Client
from errors import HandshakeError, SocketClosed


class AsyncConnection(Connection):
    def __init__(self, sock):
        sock.recv_callback = self.contruct_message
        sock.recv_close_callback = self.onclose
        self.recvbuf = []
        Connection.__init__(self, sock)

    def contruct_message(self, frame):
        if isinstance(frame, ControlFrame):
            self.handle_control_frame(frame)
            return

        self.recvbuf.append(frame)

        if frame.final:
            message = self.concat_fragments(self.recvbuf)
            self.recvbuf = []
            self.onmessage(message)
        elif len(self.recvbuf) > 1 and frame.opcode != OPCODE_CONTINUATION:
            raise ValueError('expected continuation/control frame, got %s '
                             'instead' % frame)

    def send(self, message, fragment_size=None, mask=False):
        frames = list(self.message_to_frames(message, fragment_size, mask))

        for frame in frames[:-1]:
            self.sock.queue_send(frame)

        self.sock.queue_send(frames[-1], lambda: self.onsent(message))

    def send_frame(self, frame, callback):
        self.sock.queue_send(frame, callback)

    def do_async_send(self):
        self.execute_controlled(self.sock.do_async_send)

    def do_async_recv(self, bufsize):
        self.execute_controlled(self.sock.do_async_recv, bufsize)

    def execute_controlled(self, func, *args, **kwargs):
        try:
            func(*args, **kwargs)
        except (KeyboardInterrupt, SystemExit, SocketClosed):
            raise
        except Exception as e:
            self.onerror(e)
            self.onclose(None, 'error: %s' % e)

            try:
                self.sock.close()
            except socket.error:
                pass

            raise e

    def send_close_frame(self, code, reason):
        self.sock.queue_send(create_close_frame(code, reason),
                             self.shutdown_write)
        self.close_frame_sent = True

    def close(self, code=None, reason=''):
        self.send_close_frame(code, reason)

    def send_ping(self, payload=''):
        self.sock.queue_send(ControlFrame(OPCODE_PING, payload),
                             lambda: self.onping(payload))
        self.ping_payload = payload
        self.ping_sent = True

    def onsent(self, message):
        """
        Called after a message has been written.
        """
        return NotImplemented


class AsyncServer(Server):
    def __init__(self, *args, **kwargs):
        Server.__init__(self, *args, **kwargs)

        self.recvbuf_size = kwargs.get('recvbuf_size', 2048)

        self.epoll = epoll()
        self.epoll.register(self.sock.fileno(), EPOLLIN)
        self.conns = {}

    @property
    def clients(self):
        return self.conns.values()

    def remove_client(self, client, code, reason):
        self.epoll.unregister(client.fno)
        del self.conns[client.fno]
        self.onclose(client, code, reason)

    def handle_events(self):
        for fileno, event in self.epoll.poll(1):
            if fileno == self.sock.fileno():
                try:
                    sock, addr = self.sock.accept()
                except HandshakeError as e:
                    logging.error('Invalid request: %s', e.message)
                    continue

                client = AsyncClient(self, sock)
                client.fno = sock.fileno()
                sock.setblocking(0)
                self.epoll.register(client.fno, EPOLLIN)
                self.conns[client.fno] = client
                logging.debug('Registered client %s', client)

            elif event & EPOLLHUP:
                self.epoll.unregister(fileno)
                del self.conns[fileno]

            else:
                conn = self.conns[fileno]

                try:
                    if event & EPOLLOUT:
                        conn.do_async_send()
                    elif event & EPOLLIN:
                        conn.do_async_recv(self.recvbuf_size)
                except (KeyboardInterrupt, SystemExit):
                    raise
                except SocketClosed:
                    continue
                except Exception as e:
                    logging.error(format_exc(e).rstrip())
                    continue

                self.update_mask(conn)

    def run(self):
        try:
            while True:
                self.handle_events()
        except (KeyboardInterrupt, SystemExit):
            logging.info('Received interrupt, stopping server...')
        finally:
            self.epoll.unregister(self.sock.fileno())
            self.epoll.close()
            self.sock.close()

    def update_mask(self, conn):
        mask = 0

        if conn.sock.can_send():
            mask |= EPOLLOUT

        if conn.sock.can_recv():
            mask |= EPOLLIN

        self.epoll.modify(conn.sock.fileno(), mask)

    def onsent(self, client, message):
        return NotImplemented


class AsyncClient(Client, AsyncConnection):
    def __init__(self, server, sock):
        self.server = server
        AsyncConnection.__init__(self, sock)

    def send(self, message, fragment_size=None, mask=False):
        logging.debug('Enqueueing %s to %s', message, self)
        AsyncConnection.send(self, message, fragment_size, mask)
        self.server.update_mask(self)

    def onsent(self, message):
        logging.debug('Finished sending %s to %s', message, self)
        self.server.onsent(self, message)


if __name__ == '__main__':
    import sys
    port = int(sys.argv[1]) if len(sys.argv) > 1 else 8000
    AsyncServer(('', port), loglevel=logging.DEBUG).run()



import zlib

from extension import Extension
from frame import ControlFrame


class DeflateFrame(Extension):
    """
    This is an implementation of the "deflate-frame" extension, as defined by
    http://tools.ietf.org/html/draft-tyoshino-hybi-websocket-perframe-deflate-06.

    Supported parameters are:
    - max_window_size: maximum size for the LZ77 sliding window.
    - no_context_takeover: disallows usage of LZ77 sliding window from
                           previously built frames for the current frame.

    Note that the deflate and inflate hooks modify the RSV1 bit and payload of
    existing `Frame` objects.
    """
    names = ('deflate-frame', 'x-webkit-deflate-frame')
    rsv1 = True
    defaults = {
        'max_window_bits': zlib.MAX_WBITS,
        'no_context_takeover': False
    }

    compression_threshold = 20  # minimal payload size for compression

    def negotiate(self, name, params):
        if 'max_window_bits' in params:
            mwb = int(params['max_window_bits'])
            assert 8 <= mwb <= zlib.MAX_WBITS
            yield 'max_window_bits', mwb

        if 'no_context_takeover' in params:
            assert params['no_context_takeover'] is True
            yield 'no_context_takeover', True

    class Instance(Extension.Instance):
        def init(self):
            if not self.no_context_takeover:
                self.defl = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                        zlib.DEFLATED, -self.max_window_bits)
                self.dec = zlib.decompressobj(-self.max_window_bits)

        def onsend(self, frame):
            if not frame.rsv1 and not isinstance(frame, ControlFrame) and \
                   len(frame.payload) > self.extension.compression_threshold:
                deflated = self.deflate(frame.payload)

                if len(deflated) < len(frame.payload):
                    frame.rsv1 = True
                    frame.payload = deflated

        def onrecv(self, frame):
            if frame.rsv1:
                if isinstance(frame, ControlFrame):
                    raise ValueError('received compressed control frame')

                frame.rsv1 = False
                frame.payload = self.inflate(frame.payload)

        def deflate(self, data):
            if self.no_context_takeover:
                self.defl = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                        zlib.DEFLATED, -self.max_window_bits)

            compressed = self.defl.compress(data)
            compressed += self.defl.flush(zlib.Z_SYNC_FLUSH)
            assert compressed[-4:] == '\x00\x00\xff\xff'
            return compressed[:-4]

        def inflate(self, data):
            data = str(data + '\x00\x00\xff\xff')

            if self.no_context_takeover:
                self.dec = zlib.decompressobj(-self.max_window_bits)

            return self.dec.decompress(data)



import socket
import ssl

from frame import receive_frame, pop_frame, contains_frame
from handshake import ServerHandshake, ClientHandshake
from errors import SSLError


INHERITED_ATTRS = ['bind', 'close', 'listen', 'fileno', 'getpeername',
                   'getsockname', 'getsockopt', 'setsockopt', 'setblocking',
                   'settimeout', 'gettimeout', 'shutdown', 'family', 'type',
                   'proto']

class websocket(object):
    """
    Implementation of web socket, upgrades a regular TCP socket to a websocket
    using the HTTP handshakes and frame (un)packing, as specified by RFC 6455.
    The API of a websocket is identical to that of a regular socket, as
    illustrated by the examples below.

    Server example:
    >>> import wspy, socket
    >>> sock = wspy.websocket()
    >>> sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    >>> sock.bind(('', 8000))
    >>> sock.listen(5)

    >>> client = sock.accept()
    >>> client.send(wspy.Frame(wspy.OPCODE_TEXT, 'Hello, Client!'))
    >>> frame = client.recv()

    Client example:
    >>> import wspy
    >>> sock = wspy.websocket(location='/my/path')
    >>> sock.connect(('', 8000))
    >>> sock.send(wspy.Frame(wspy.OPCODE_TEXT, 'Hello, Server!'))
    """
    def __init__(self, sock=None, origin=None, protocols=[], extensions=[],
                 location='/', trusted_origins=[], locations=[], auth=None,
                 recv_callback=None, sfamily=socket.AF_INET, sproto=0):
        """
        Create a regular TCP socket of family `family` and protocol

        `sock` is an optional regular TCP socket to be used for sending binary
        data. If not specified, a new socket is created.

        `origin` (for client sockets) is the value for the "Origin" header sent
        in a client handshake .

        `protocols` is a list of supported protocol names.

        `extensions` (for server sockets) is a list of supported extensions
        (`Extension` instances).

        `location` (for client sockets) is optional, used to request a
        particular resource in the HTTP handshake. In a URL, this would show as
        ws://host[:port]/<location>. Use this when the server serves multiple
        resources (see `locations`).

        `trusted_origins` (for server sockets) is a list of expected values
        for the "Origin" header sent by a client. If the received Origin header
        has value not in this list, a HandshakeError is raised. If the list is
        empty (default), all origins are excepted.

        `locations` (for server sockets) is an optional list of resources
        serverd by this server. If specified (without trailing slashes), these
        are used to verify the resource location requested by a client. The
        requested location may be used to distinquish different services in a
        server implementation.

        `auth` is optional, used for HTTP Basic or Digest authentication during
        the handshake. It must be specified as a (username, password) tuple.

        `recv_callback` is the callback for received frames in asynchronous
        sockets. Use in conjunction with setblocking(0). The callback itself
        may for example change the recv_callback attribute to change the
        behaviour for the next received message. Can be set when calling
        `queue_send`.

        `sfamily` and `sproto` are used for the regular socket constructor.
        """
        self.protocols = protocols
        self.extensions = extensions
        self.extension_instances = []
        self.origin = origin
        self.location = location
        self.trusted_origins = trusted_origins
        self.locations = locations
        self.auth = auth

        self.secure = False

        self.handshake_sent = False

        self.sendbuf_frames = []
        self.sendbuf = ''
        self.recvbuf = ''
        self.recv_callback = recv_callback

        self.sock = sock or socket.socket(sfamily, socket.SOCK_STREAM, sproto)

    def __getattr__(self, name):
        if name in INHERITED_ATTRS:
            return getattr(self.sock, name)

        raise AttributeError("'%s' has no attribute '%s'"
                             % (self.__class__.__name__, name))

    def accept(self):
        """
        Equivalent to socket.accept(), but transforms the socket into a
        websocket instance and sends a server handshake (after receiving a
        client handshake). Note that the handshake may raise a HandshakeError
        exception.
        """
        sock, address = self.sock.accept()
        wsock = websocket(sock)
        wsock.secure = self.secure
        ServerHandshake(wsock).perform(self)
        wsock.handshake_sent = True
        return wsock, address

    def connect(self, address):
        """
        Equivalent to socket.connect(), but sends an client handshake request
        after connecting.

        `address` is a (host, port) tuple of the server to connect to.
        """
        self.sock.connect(address)
        ClientHandshake(self).perform()
        self.handshake_sent = True

    def apply_send_hooks(self, frame, before_fragmentation):
        for inst in self.extension_instances:
            if inst.extension.before_fragmentation == before_fragmentation:
                frame = inst.handle_send(frame)

        return frame

    def apply_recv_hooks(self, frame, before_fragmentation):
        for inst in reversed(self.extension_instances):
            if inst.extension.before_fragmentation == before_fragmentation:
                frame = inst.handle_recv(frame)

        return frame

    def send(self, *args):
        """
        Send a number of frames.
        """
        for frame in args:
            self.sock.sendall(self.apply_send_hooks(frame, False).pack())

    def recv(self):
        """
        Receive a single frames. This can be either a data frame or a control
        frame.
        """
        return self.apply_recv_hooks(receive_frame(self.sock), False)

    def recvn(self, n):
        """
        Receive exactly `n` frames. These can be either data frames or control
        frames, or a combination of both.
        """
        return [self.recv() for i in xrange(n)]

    def queue_send(self, frame, callback=None, recv_callback=None):
        """
        Enqueue `frame` to the send buffer so that it is send on the next
        `do_async_send`. `callback` is an optional callable to call when the
        frame has been fully written. `recv_callback` is an optional callable
        to quickly set the `recv_callback` attribute to.
        """
        frame = self.apply_send_hooks(frame, False)
        self.sendbuf += frame.pack()
        self.sendbuf_frames.append([frame, len(self.sendbuf), callback])

        if recv_callback:
            self.recv_callback = recv_callback

    def do_async_send(self):
        """
        Send any queued data. This function should only be called after a write
        event on a file descriptor.
        """
        assert len(self.sendbuf)

        nwritten = self.sock.send(self.sendbuf)
        nframes = 0

        for entry in self.sendbuf_frames:
            frame, offset, callback = entry

            if offset <= nwritten:
                nframes += 1

                if callback:
                    callback()
            else:
                entry[1] -= nwritten

        self.sendbuf = self.sendbuf[nwritten:]
        self.sendbuf_frames = self.sendbuf_frames[nframes:]

    def do_async_recv(self, bufsize):
        """
        Receive any completed frames from the socket. This function should only
        be called after a read event on a file descriptor.
        """
        data = self.sock.recv(bufsize)

        if len(data) == 0:
            raise socket.error('no data to receive')

        self.recvbuf += data

        while contains_frame(self.recvbuf):
            frame, self.recvbuf = pop_frame(self.recvbuf)
            frame = self.apply_recv_hooks(frame, False)

            if not self.recv_callback:
                raise ValueError('no callback installed for %s' % frame)

            self.recv_callback(frame)

    def can_send(self):
        return len(self.sendbuf) > 0

    def can_recv(self):
        return self.recv_callback is not None

    def enable_ssl(self, *args, **kwargs):
        """
        Transforms the regular socket.socket to an ssl.SSLSocket for secure
        connections. Any arguments are passed to ssl.wrap_socket:
        http://docs.python.org/dev/library/ssl.html#ssl.wrap_socket
        """
        if self.handshake_sent:
            raise SSLError('can only enable SSL before handshake')

        self.secure = True
        self.sock = ssl.wrap_socket(self.sock, *args, **kwargs)
